/*
 * Observer.java
 *
 * Gary Givental
 * John Gotts
 * Saul Rodriguez
 *
 * EECS 481
 *
 * Revision History:
 *
 * 09/09/97 - original implementation.
 * 11/28/97 - added beginnings of networking functionality.
 *
 */

import java.util.*;
import java.awt.*;

/**
 * Implements an observer.  The observer watches the progress of the game and
 * reports game statistics, i.e. what cards have been played by what player,
 * who picked up the trump card, etc.
 *
 * @version     1.0
 * @author      Gary Givental
 * @author      John Gotts
 * @author      Saul Rodriguez
 */
public class Observer {

public static final int SOUTH = 0;
public static final int WEST = 1;
public static final int NORTH = 2;
public static final int EAST = 3;
public static final int CENTER = 4;
protected Network network;
protected Euchre euchre;
protected Vector record = new Vector();	// Card play history.
protected int playNum; 			// Number of cards played.
protected Player[] players = new Player[4];
protected Card[] playedCards = new Card[4];
protected int currentTrick;		// The current trick number (1-5).
protected int dealer;			// The dealer's position.
protected int currentPosition;		// Thr current position.
protected int leader;			// The leader.
protected int playerNumber = 0;		// Beware, playerNumber is numbered
					// from 1 to 4.  This is only used by
					// the network functionality.
protected int trumpCaller;		// The player that called trump.
protected int aloneCaller = -1;		// The player that requested to go
					// alone.
protected int trickWinner;
protected int[] tricks = new int[4];	// Number of tricks each player has won.
protected boolean master = false;	// Whether the table belongs to us.
					// If so, we have to network deal and
					// establish the trump card in a
					// networked manner.  If not, we listen
					// to these sorts of messages.  Only
					// relevant to a networked game.
protected boolean inProgress = false;
protected boolean pickedUpTrump;	// Whether the trump card was actually
					// picked up.
protected Card networkCard;
protected boolean networkChoice;	// Used for network trump.
protected boolean waitingForHumanDiscardedCard = false;
protected boolean waitingForHumanPlayedCard = false;
protected boolean waitingForNetworkCalled = false;
protected boolean waitingForNetworkDiscardedCard = false;
protected boolean waitingForNetworkCalledSuit = false;
protected boolean waitingForNetworkPlayedCard = false;

// Might be useless now.
protected int firstGoAroundCount;	// Used for network trump.
protected int secondGoAroundCount;	// Used for network trump.
protected boolean choiceMade;		// Used for network trump.

/**
 * Constructs the observer.
 *
 * @param euchre	the card game.
 */
public Observer (Euchre euchre) {

	this.euchre = euchre;

	// Don't rotate here.
	players[SOUTH] = null;
	players[WEST] = null;
	players[NORTH] = null;
	players[EAST] = null;

	resetObserver();
}

/**
 * Returns vector containing played cards in current trick.
 *
 * @return 		Cards played on trick.
 */
public Vector getPlayedCards ( ) {

	Vector vector = new Vector();
	int i;
	
	for (i = leader; playedCards[i] != null; i = rotate(i)) {
		vector.addElement(playedCards[i]);
	}
	
	return vector;
}

/**
 * Determines if a card has been played in the hand yet.
 *
 * @return 	True if next card will be first played in hand.
 */
public boolean firstPlay ( ) {

	if (playedCards[leader] == null)
		return true;
	else
		return false;
}

/**
 * Returns card that was led on the current trick.
 *
 * @return 		Card led on trick.
 */
public Card getLedCard ( ) {

	return playedCards[leader];
}

/**
 * Returns suit that was led on the current trick. Returns Card.NONE if no card
 * has been led yet.
 *
 * @return			Suit of led card.
 */
public int getLedSuit ( ) {

	Card card;
	
	card = getLedCard ( );
	
	if (card == null)
		return Card.NONE;
	
	return card.getSuit();
}

public void playedTrick () {

	if (network != null)
		processOutgoingText(new String("U playedTrick"));
}

public void updatedScore (int position) {

	int actualPlayerNumber;

	actualPlayerNumber = playerNumber + position;
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null) {
		if (position == NORTH || position == SOUTH)
			processOutgoingText(new String("U " + actualPlayerNumber + " updatedScore " + euchre.getTable().getOurScore()));
		else
			processOutgoingText(new String("U " + actualPlayerNumber + " updatedScore " + euchre.getTable().getTheirScore()));
	}
}

public void updatedTricks (int position, int tricks) {

	int actualPlayerNumber;

	actualPlayerNumber = playerNumber + position;
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " updatedTricks " + tricks));
}

public void assignedTrickWinner (Player player) {

	int actualPlayerNumber;

	euchre.getGUI().blockAdvise("Winner", euchre.getGUI().getPlayerName(player.getPosition()) + " wins.");
	actualPlayerNumber = playerNumber + player.getPosition();
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " assignedTrickWinner"));
}

/**
 * Notifies everyone that the player played the specified card.  As a side
 * effect, adds to the record of played cards.
 *
 * @param player	the player in question.
 * @param card		the card in question.
 */
public void playedCard (Player player, Card card) {

	int actualPlayerNumber, i;

	// Fill in the playedCards vector.
	for (i = 0; i < 4; i++)
		if (player.equals(players[i])) {
			playedCards[i] = card;
			break;
		}
	actualPlayerNumber = playerNumber + player.getPosition();
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " playedCard "+ card.getValue() + " " + card.getSuit() + " " + card.getSubRank()));
}

public void discardedCard (Player player, Card card) {

	int actualPlayerNumber;

	actualPlayerNumber = playerNumber + player.getPosition();
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " discardedCard "+ card.getValue() + " " + card.getSuit() + " " + card.getSubRank()));
}

/**
 * Notifies everyone that the player requested the dealer to pick up the trump
 * card.
 *
 * @param player	the player in question.
 */
public void called (Player player) {

	int actualPlayerNumber;

	trumpCaller = player.getPosition();
	euchre.getGUI().updateTrumpCaller(trumpCaller);
	euchre.getGUI().updateTrump(euchre.getTable().getTrumpCard().getDescription());
	actualPlayerNumber = playerNumber + player.getPosition();
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " called"));
}

/**
 * Notifies everyone that the player called a suit.
 *
 * @param player	the player in question.
 */
public void calledSuit (Player player, int suit) {

	int actualPlayerNumber;

	euchre.getTable().setTrumpSuit(suit);
	trumpCaller = player.getPosition();
	System.out.println("Updating trump caller and trump.");
	euchre.getGUI().updateTrumpCaller(trumpCaller);
	euchre.getGUI().updateTrump(Card.getSuitDescription(euchre.getTable().getTrumpSuit()));
	actualPlayerNumber = playerNumber + player.getPosition();
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " calledSuit " + euchre.getTable().getTrumpSuit()));
}

public void assignedLeader (Player player) {

	int actualPlayerNumber;

	leader = player.getPosition();
	euchre.getGUI().updateLeader(leader);
	actualPlayerNumber = playerNumber + player.getPosition();
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " assignedLeader"));
}

public void assignedDealer (Player player) {

	int actualPlayerNumber;

	dealer = player.getPosition();
	euchre.getGUI().updateDealer(dealer);
	actualPlayerNumber = playerNumber + player.getPosition();
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " assignedDealer"));
}

public void passed (Player player) {

	int actualPlayerNumber;
	int position = player.getPosition();

	if (position != SOUTH)
		euchre.getGUI().advise("Passed", new String(euchre.getGUI().getPlayerName(player.getPosition()) + " passed."));
	if (position == dealer) {
		euchre.getTable().setPassed(true);
		euchre.getGUI().removeCard(euchre.getTable().getTrumpCard(), CENTER);
	}
	actualPlayerNumber = playerNumber + position;
	if (actualPlayerNumber > 4)
		actualPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + actualPlayerNumber + " passed"));
}

public void trumpCardDealt (Card trumpCard) {

	euchre.getTable().setTrumpCard(trumpCard);
	if (network != null && master)
		processOutgoingText(new String("U trumpCardDealt " + trumpCard.getValue() + " " + trumpCard.getSuit() + " " + trumpCard.getSubRank()));
}

public void cardDealt (Player player, Card card) {

	int recipientPlayerNumber;

	recipientPlayerNumber = playerNumber + player.getPosition();
	if (recipientPlayerNumber > 4)
		recipientPlayerNumber -= 4;
	if (network != null)
		processOutgoingText(new String("U " + recipientPlayerNumber + " cardDealt " + card.getValue() + " " + card.getSuit() + " " + card.getSubRank()));
}

/**
 * Resets the observer for a new hand.
 */
public void newHand () {

	tricks[SOUTH] = 0;
	updatedTricks(SOUTH, 0);
	tricks[EAST] = 0;
	updatedTricks(EAST, 0);
	tricks[NORTH] = 0;
	updatedTricks(NORTH, 0);
	tricks[WEST] = 0;
	updatedTricks(WEST, 0);
	euchre.getGUI().updateOurTricks(0);
	euchre.getGUI().updateTheirTricks(0);
	record = new Vector();
	euchre.getTable().reset();
	euchre.getDeck().reset();
	trumpCaller = -1;
	if (network != null)
		processOutgoingText(new String("U newHand"));
}

/**
 * Returns whether the game is in progress.
 *
 * @return	whether the game is in progress.
 */
public boolean isInProgress () {

	return inProgress;
}

/**
 * Deals a hand.
 */
public void deal () {

	Card card;
	Player player;
	NetworkPlayer networkPlayer;
	HumanPlayer humanPlayer;
	AIPlayer aiPlayer;
	Vector cards;
	int i, j;

	System.out.println("Deal...");

	// For now, deal the player five cards in a row.  Later, to beautify
	// this program, deal out cards 3 2 3 2 2 3 2 3 or whatever.
	for (i = 0; i < 4; i++) {
		player = players[i];
		cards = new Vector();
		if (player instanceof AIPlayer) {
			aiPlayer = (AIPlayer) player;
			for (j = 0; j < 5; j++) {
				card = euchre.getDeck().drawCard();
				cards.addElement(card);
				cardDealt(aiPlayer, card);
			}
			aiPlayer.addCards(cards);
		} else if (player instanceof HumanPlayer) {
			humanPlayer = (HumanPlayer) player;
			for (j = 0; j < 5; j++) {
				card = euchre.getDeck().drawCard();
        			card.setFaceUp(true);
				euchre.getGUI().addCard(card, SOUTH, CENTER);
				cards.addElement(card);
				cardDealt(humanPlayer, card);
			}
			humanPlayer.addCards(cards);
		} else if (player instanceof NetworkPlayer) {
			networkPlayer = (NetworkPlayer) player;
			for (j = 0; j < 5; j++) {
				card = euchre.getDeck().drawCard();
				cards.addElement(card);
				cardDealt(networkPlayer, card);
			}
			networkPlayer.addCards(cards);
		} else {
			System.out.println("Observer.java: Invalid player type.");
			System.exit(0);
		}
	}
	// Deal the trump card.
	card = euchre.getDeck().drawCard();
	card.setFaceUp(true);
	euchre.getGUI().addCard(card, CENTER, CENTER);
	trumpCardDealt(card);
}

/**
 * Determines who won and assigns appropriate scores.
 */
public void processScores () {

	// Note: If a player goes alone, his partner will automatically have
	// no tricks.
	if (tricks[SOUTH] + tricks [NORTH] > tricks[WEST] + tricks[EAST]) {
		if (tricks[SOUTH] + tricks[NORTH] == 5) {
			if (aloneCaller == SOUTH || aloneCaller == NORTH)
				euchre.getTable().incrementOurScore (4);
			else
				euchre.getTable().incrementOurScore (2);
		} else {
			if (trumpCaller == EAST || trumpCaller == WEST)
				euchre.getTable().incrementOurScore (2);
			else
				euchre.getTable().incrementOurScore (1);
		}
		euchre.getGUI().updateOurScore(euchre.getTable().getOurScore());
		updatedScore(SOUTH);
	} else {
		if (tricks[EAST] + tricks[WEST] == 5) {
			if (aloneCaller == EAST || aloneCaller == WEST)
				euchre.getTable().incrementTheirScore (4);
			else
				euchre.getTable().incrementTheirScore (2);
		} else {
			if (trumpCaller == NORTH || trumpCaller == SOUTH)
				euchre.getTable().incrementTheirScore (2);
			else
				euchre.getTable().incrementTheirScore (1);
		}
		euchre.getGUI().updateTheirScore(euchre.getTable().getTheirScore());
		updatedScore(WEST);
	}
}

/**
 * Returns our actual player number from 1 to 4.  Only used for the networking
 * functionality.
 *
 * @return	our actual player number from 1 to 4.
 */
public int getPlayerNumber () {

	return playerNumber;
}

/**
 * Processes incoming text from the network.
 *
 * @param text	the incoming text message.
 */
public void processIncomingText(String text) {

	StringTokenizer stringTokenizer = new StringTokenizer(text);
	String command = null;
	String token = null;
	String textString;
	String value = null;
	String name;
	int source, position, suit;
	int destination = -1;

	// Process a particular command.
	if (stringTokenizer.hasMoreTokens())
		token = stringTokenizer.nextToken();
	else {
		System.out.println("Observer.java: Got a null line.");
		System.exit(0);
	}
	try {
		source = Integer.parseInt(token);
	} catch (NumberFormatException exception) {
		// No source for some commands.
		command = token;
		source = 0;
	} catch (Exception exception) {
		System.out.println("Unknown exception: " + exception + ".");
		return;
	}
	if (command == null) {
		if (stringTokenizer.hasMoreTokens())
			command = stringTokenizer.nextToken();
		else {
			System.out.println("Observer.java: A source with no command.");
			System.exit(0);
		}
	}
	// See if there is a destination.
	if (stringTokenizer.hasMoreTokens())
		token = stringTokenizer.nextToken();
	else
		token = null;
	if (token != null) {
		try {
			destination = Integer.parseInt(token);
		} catch (NumberFormatException exception) {
			// No destination for some commands.
			destination = -1;
		} catch (Exception exception) {
			System.out.println("Unknown exception: " + exception + ".");
			System.exit(0);
		}
	} else
		destination = -1;
	// A player joins.
	if (command.compareTo("N") == 0) {
		if (source < 1 || source > 4) {
			System.out.println("Observer.java: New sent without or with invalid player number.");
			System.exit(0);
		}
		syncUpNewPlayer(source);
	// My name is...
	} else if (command.compareTo("J") == 0) {
		if (source < 1 || source > 4) {
			System.out.println("Observer.java: Name sent without or with invalid player number.");
			System.exit(0);
		}
		position = source - playerNumber;
		if (position < 0)
			position += 4;
		setPlayerName(text.substring(4), position);
	// Someone says...
	} else if (command.compareTo("S") == 0) {
		if (source < 1 || source > 4) {
			System.out.println("Observer.java: Chat message sent without or with invalid player number.");
			System.exit(0);
		}
		if (playerNumber != 0) {
			position = source - playerNumber;
			if (position < 0)
				position += 4;
		} else
			position = 0;
		name = euchre.getGUI().getPlayerName(position);
		// Ignore chat messages before the user sets up the game.
		if (name.length() > 0) {
			textString = new String("<" + name + "> " + text.substring(4));
			euchre.getConsole().addText(textString);
		}
	// A new game starts.
	} else if (command.compareTo("G") == 0) {
		System.out.println("Received a G to start the game.");
		// Don't need to notify ourselves of a new game; we already
		// know.
		if (source != playerNumber) {
			inProgress = true;
			playGame(false);
		}
		else
			System.out.println("Received my own G.  Not doing anything");
	// A control message was sent.
	} else if (command.compareTo("C") == 0) {
		if (destination == -1) {
			System.out.println("Observer.java: Control command with no destination.");
			System.exit(0);
		}
		position = destination - playerNumber;
		if (position < 0)
			position += 4;
		System.out.println("Got a C message for " + position + ".");
		if (position == SOUTH && ! master)
			processControlMessage(text.substring(6));
		else
			System.out.println("Not for me: " + text);
	// A game update message was sent.
	} else if (command.compareTo("U") == 0) {
		processUpdateMessage(text.substring(4));
	// A play/call message was sent.
	} else if (command.compareTo("P") == 0) {
	// A pass message was sent.
	} else if (command.compareTo("W") == 0) {
	// A local message was sent.
	} else if (command.compareTo("L") == 0) {
		euchre.getConsole().addText(text.substring(4));
	// Error.
	} else if (command.compareTo("E") == 0) {
		position = source - playerNumber;
		if (position < 0)
			position += 4;
		euchre.getGUI().advise("Player Disconnected", new String(euchre.getGUI().getPlayerName(position) + " has left the game due to an error."));
		unsetPlayerName(position);
	// Quit.
	} else if (command.compareTo("Q") == 0) {
		position = source - playerNumber;
		if (position < 0)
			position += 4;
		euchre.getGUI().advise("Player Disconnected", new String(euchre.getGUI().getPlayerName(position) + " has left the game."));
		unsetPlayerName(position);
	} else {
		textString = new String("Received: " + text);
		euchre.getConsole().addText(textString);
	}
}

public void processOutgoingText(String text) {

	StringTokenizer stringTokenizer = new StringTokenizer(text);

	if (network != null)
		network.sendMessage(text);
	else
		// When the network's not up, act like a loopback.
		processIncomingText("1 " + text);

}

/**
 * Dumps debugging information to the console.  All text messages will start
 * with "% ".
 *
 * @param text	the debugging text message.
 */
public void debug (String text) {

	System.out.println("Observer.java: debug() called with " + text + ".");
}

/**
 * Adds an initial human player.
 *
 * @param position	the player's position.
 */
public void addHumanPlayer (int position) {

	players[position] = new HumanPlayer(euchre, this, position);
}

/**
 * Adds an initial AI player.
 *
 * @param position	the player's position.
 */
public void addAIPlayer (int position) {

	players[position] = new AIPlayer(euchre, this, position);
}

/**
 * Slave network method.
 */
public void processUpdateMessage (String message) {

	Vector cards = new Vector();
	Card card;
	StringTokenizer stringTokenizer = new StringTokenizer(message);
	String token = null;
	String value = null;
	int i, suit, subRank, source, position, score;
	int realPosition;

	System.out.println("Update message: " + message + ".");
	if (stringTokenizer.hasMoreTokens())
		token = stringTokenizer.nextToken();
	else {
		System.out.println("Observer.java: Missing update message command.");
		System.exit(0);
	}
	if (token.compareTo("trumpCardDealt") == 0) {
		if (! master) {
			card = createCard(message.substring(15));
			card.setFaceUp(true);
			euchre.getGUI().addCard(card, CENTER, CENTER);
			euchre.getTable().setTrumpCard(card);
		}
	} else if (token.compareTo("newHand") == 0) {
		if (! master) {
			tricks[SOUTH] = 0;
			updatedTricks(SOUTH, 0);
			tricks[EAST] = 0;
			updatedTricks(EAST, 0);
			tricks[NORTH] = 0;
			updatedTricks(NORTH, 0);
			tricks[WEST] = 0;
			updatedTricks(WEST, 0);
			euchre.getGUI().updateOurTricks(0);
			euchre.getGUI().updateTheirTricks(0);
			record = new Vector();
			euchre.getTable().reset();
			euchre.getDeck().reset();
			trumpCaller = -1;
		}
	} else if (token.compareTo("playedTrick") == 0) {
		if (! master)
			for (i = 0; i < 4; i++) {
				euchre.getGUI().removeCard(playedCards[i], CENTER);
				playedCards[i] = null;
			}
	} else {
		// trumpCardDealt messages are the only ones without targets.
		source = Integer.parseInt(token);
		if (stringTokenizer.hasMoreTokens())
			token = stringTokenizer.nextToken();
		else {
			System.out.println("Observer.java: Missing update message command.");
			System.exit(0);
		}
		position = source - playerNumber;
		if (position < 0)
			position += 4;
		if (token.compareTo("cardDealt") == 0) {
			if (! master) {
				// Note: Add two to account for the target.
				card = createCard(message.substring(12));
				if (players[position] instanceof HumanPlayer) {
					card.setFaceUp(true);
					euchre.getGUI().addCard(card, SOUTH, CENTER);
				}
				System.out.println("Added card " + card.getDescription() + " for player " + position + ".");
				cards.addElement(card);
				players[position].addCards(cards);
			}
		} else if (token.compareTo("assignedTrickWinner") == 0) {
			if (! master)
				euchre.getGUI().blockAdvise("Winner", euchre.getGUI().getPlayerName(position) + " wins.");
		} else if (token.compareTo("playedCard") == 0) {
			if (waitingForNetworkPlayedCard) {
				if (position != currentPosition) return;
				// Note: Add two to account for the target.
				card = findLocalCard(players[position], createCard(message.substring(13)));
				if (card == null) {
					System.out.println("That player didn't have that card.");
					System.exit(0);
				}
				playedCards[position] = card;
				card.setFaceUp(true);
				euchre.getGUI().addCard(card, Observer.CENTER, position);
				realPosition = position + 1;
				System.out.println("Removing player " + realPosition + "'s played card.");
				players[position].getHand().removeCard(card);
				networkCard = card;
				playGame(false);
			} else if (! master && position != SOUTH) {
				// Note: Add two to account for the target.
				System.out.println(message.substring(13));
				card = findLocalCard(players[position], createCard(message.substring(13)));
				if (card == null) {
					System.out.println("That player didn't have that card.");
					System.exit(0);
				}
				playedCards[position] = card;
				card.setFaceUp(true);
				euchre.getGUI().addCard(card, Observer.CENTER, position);
				realPosition = position + 1;
				System.out.println("Removing player " + realPosition + "'s played card.");
				players[position].getHand().removeCard(card);
			}
		} else if (token.compareTo("called") == 0) {
			// Ignore extraneous messages.
//			if (position != currentPosition) return;
			if (waitingForNetworkCalled) {
				euchre.getGUI().advise("Called", new String(euchre.getGUI().getPlayerName(position) + " calls it up."));
				trumpCaller = position;
				euchre.getGUI().updateTrumpCaller(trumpCaller);
				euchre.getGUI().updateTrump(euchre.getTable().getTrumpCard().getDescription());
				networkChoice = true;
				playGame(false);
			} else if (! master) {
				euchre.getGUI().blockAdvise("Called", new String(euchre.getGUI().getPlayerName(position) + " calls it up."));
				trumpCaller = position;
				euchre.getGUI().updateTrumpCaller(trumpCaller);
				euchre.getGUI().updateTrump(euchre.getTable().getTrumpCard().getDescription());
				players[dealer].getHand().addCard(euchre.getTable().getTrumpCard());
				euchre.getGUI().removeCard(euchre.getTable().getTrumpCard(), CENTER);
			}
		} else if (token.compareTo("calledSuit") == 0) {
			// Ignore extraneous messages.
//			if (position != currentPosition) return;
			if (waitingForNetworkCalledSuit) {
				if (stringTokenizer.hasMoreTokens())
					token = stringTokenizer.nextToken();
				else {
					System.out.println("Observer.java: calledSuit without suit.");
					System.exit(0);
				}
				suit = Integer.parseInt(token);
				euchre.getTable().setTrumpSuit(suit);
				trumpCaller = position;
				euchre.getGUI().updateTrumpCaller(trumpCaller);
				euchre.getGUI().updateTrump(Card.getSuitDescription(euchre.getTable().getTrumpSuit()));
				networkChoice = true;
				playGame(false);
			} else if (! master) {
				if (stringTokenizer.hasMoreTokens())
					token = stringTokenizer.nextToken();
				else {
					System.out.println("Observer.java: calledSuit without suit.");
					System.exit(0);
				}
				suit = Integer.parseInt(token);
				euchre.getTable().setTrumpSuit(suit);
				euchre.getGUI().blockAdvise("Call it Up", euchre.getGUI().getPlayerName(position) + " calls " + Card.getSuitDescription(suit));
				trumpCaller = position;
				System.out.println("Updating trump caller and trump.");
				euchre.getGUI().updateTrumpCaller(trumpCaller);
				euchre.getGUI().updateTrump(Card.getSuitDescription(euchre.getTable().getTrumpSuit()));
			}
		} else if (token.compareTo("assignedLeader") == 0) {
			leader = position;
			euchre.getGUI().updateLeader(leader);
		} else if (token.compareTo("assignedDealer") == 0) {
			if (! master) {
				dealer = position;
				euchre.getGUI().updateDealer(dealer);
				players[dealer].setDealer(true);
			}
		} else if (token.compareTo("passed") == 0) {
			if (position != SOUTH && ((! master) || players[position] instanceof NetworkPlayer))
				euchre.getGUI().advise("Passed", new String(euchre.getGUI().getPlayerName(position) + " passed."));
			if (position == dealer && ! euchre.getTable().isPassed()) {
				euchre.getTable().setPassed(true);
				euchre.getGUI().removeCard(euchre.getTable().getTrumpCard(), CENTER);
			}
			if (waitingForNetworkCalled) {
				if (position != currentPosition) return;
				networkChoice = false;
				playGame(false);
			} else if (waitingForNetworkCalledSuit) {
				if (position != currentPosition) return;
				networkChoice = false;
				playGame(false);
			}
		} else if (token.compareTo("discardedCard") == 0) {
			if (waitingForNetworkDiscardedCard) {
				System.out.println("Removing card from network player's hand.");
				// Note: Add two to account for the target.
				card = findLocalCard(players[position], createCard(message.substring(16)));
				if (card == null) {
					System.out.println("That player didn't have that card.");
					System.exit(0);
				}
				networkCard = card;
				playGame(false);
			} else if (! master) {
				card = findLocalCard(players[position], createCard(message.substring(16)));
				if (card == null) {
					System.out.println("That player didn't have that card.");
					System.exit(0);
				}
				realPosition = position + 1;
				System.out.println("Removing player " + realPosition + "'s discard.");
				players[position].getHand().removeCard(card);
			}
		} else if (token.compareTo("updatedTricks") == 0) {
			if (! master) {
				if (stringTokenizer.hasMoreTokens())
					token = stringTokenizer.nextToken();
				else {
					System.out.println("No score.");
					System.exit(0);
				}
				score = Integer.parseInt(token);
				tricks[position] = score;
				if (position == NORTH || position == SOUTH)
					euchre.getGUI().updateOurTricks(tricks[SOUTH] + tricks[NORTH]);
				else
					euchre.getGUI().updateTheirTricks(tricks[EAST] + tricks[WEST]);
			}
		} else if (token.compareTo("updatedScore") == 0) {
			if (! master) {
				if (stringTokenizer.hasMoreTokens())
					token = stringTokenizer.nextToken();
				else {
					System.out.println("No score.");
					System.exit(0);
				}
				score = Integer.parseInt(token);
				if (position == NORTH || position == SOUTH) {
					euchre.getTable().setOurScore(score);
					euchre.getGUI().updateOurScore(score);
				} else {
					euchre.getTable().setTheirScore(score);
					euchre.getGUI().updateTheirScore(score);
				}
			}
		}
	}
}

/**
 * Slave network method.
 */
public void processControlMessage (String message) {

	Card card;
	Vector cards = new Vector();
	HumanPlayer humanPlayer;
	StringTokenizer stringTokenizer = new StringTokenizer(message);
	String token = null;
	String value = null;
	int suit, subRank;
	boolean choice;

	System.out.println("Control message: " + message + ".");
	if (stringTokenizer.hasMoreTokens())
		token = stringTokenizer.nextToken();
	else {
		System.out.println("Observer.java: Missing control message command.");
		System.exit(0);
	}
	if (token.compareTo("callItUp") == 0) {
		humanPlayer = (HumanPlayer) players[SOUTH];
		choice = humanPlayer.callItUp();
		System.out.println("Got choice: " + choice + ".");
	} else if (token.compareTo("pickUp") == 0) {
		humanPlayer = (HumanPlayer) players[SOUTH];
		humanPlayer.pickUpTrump();
		euchre.getGUI().advise("Discard", "Please discard a card.");
	} else if (token.compareTo("callSuit") == 0) {
		humanPlayer = (HumanPlayer) players[SOUTH];
		choice = humanPlayer.callItUp();
		System.out.println("Got choice: " + choice + ".");
	} else if (token.compareTo("playCard") == 0) {
		euchre.getGUI().advise("Play Card", "Please play a card.");
		waitingForHumanPlayedCard = true;
	}
}

public Card findLocalCard (Player player, Card remoteCard) {

	Enumeration e;
	Card card;

	for (e = player.getCards().elements(); e.hasMoreElements();) {
		card = (Card) e.nextElement();
		if (remoteCard.getValue().compareTo(card.getValue()) == 0 && remoteCard.getSuit() == card.getSuit()) {
			System.out.println("Found match.");
			return card;
		}
	}
	return null;
}

public Card createCard (String message) {

	StringTokenizer stringTokenizer = new StringTokenizer(message);
	String token = null;
	String value = null;
	int suit, subRank;

	if (stringTokenizer.hasMoreTokens())
		value = stringTokenizer.nextToken();
	else {
		System.out.println("Observer.java: Missing value.");
		System.exit(0);
	}
	if (stringTokenizer.hasMoreTokens())
		token = stringTokenizer.nextToken();
	else {
		System.out.println("Observer.java: Missing suit.");
		System.exit(0);
	}
	suit = Integer.parseInt(token);
	if (stringTokenizer.hasMoreTokens())
		token = stringTokenizer.nextToken();
	else {
		System.out.println("Observer.java: Missing subrank.");
		System.exit(0);
	}
	subRank = Integer.parseInt(token);
	return new Card(suit, value, subRank);
}

public void syncUpNewPlayer (int playerNumber) {

	int position;

	if (inProgress) {
		// If we're the master:
		// Before the AI players are replaced below, copy their hand
		// information to the new network players.
		// Give them our table information, dealer, number of tricks,
		// etc.
		System.out.println("Adding players while a game is in progress currently does not work.");
		return;
	}
	if (this.playerNumber == 0) {
		// Set up the existing players.
		this.playerNumber = playerNumber;
		// We're the first player to join the networked game.
		if (playerNumber == 1)
			master = true;
		// We're connecting to a game with other players.
		else if (playerNumber == 2)
			players[EAST] = new NetworkPlayer(euchre, this, EAST);
		else if (playerNumber == 3) {
			players[NORTH] = new NetworkPlayer(euchre, this, NORTH);
			players[EAST] = new NetworkPlayer(euchre, this, EAST);
		} else if (playerNumber == 4) {
			players[WEST] = new NetworkPlayer(euchre, this, WEST);
			players[NORTH] = new NetworkPlayer(euchre, this, NORTH);
			players[EAST] = new NetworkPlayer(euchre, this, EAST);
		} else {
			System.out.println("Observer.java: Got an extraneous player.");
			System.exit(0);
		}
	// A new player joins the network.
	} else {
		position = playerNumber - this.playerNumber;
		if (position < 0)
			position += 4;
		players[position] = new NetworkPlayer(euchre, this, position);
	}
	// Tell everyone my name.
	processOutgoingText(new String("J " + euchre.getGUI().getPlayerName(SOUTH)));
}

/**
 * Sets a player's name.
 *
 * @param playerName	the player's name.
 * @param position	the player's position.
 */
public void setPlayerName (String playerName, int position) {

	if (position != SOUTH) {
		euchre.getGUI().modifyPlayer(playerName, position, true);
		euchre.getGUI().pack();
		euchre.getGUI().advise("New Player", new String(playerName + " has joined the game."));
	}
}

/**
 * Unsets a player's name (used for players that leave the game.
 *
 * @param position	the player's position.
 */
public void unsetPlayerName (int position) {

	// META: If we're the master, update the AI player with the old
	// player's information.
	euchre.getGUI().modifyPlayer("Computer", position, false);
	euchre.getGUI().pack();
}

/**
 * Initializes the variables for a new game.
 */
public void newGame () {

	euchre.getTable().resetScores();
	resetDealer();
	resetObserver();
}

/**
 * Unconditionally resets the observer variables.
 */
public void resetObserver () {

	playNum = 0;
	record.removeAllElements();
}

/**
 * Unconditionally sets the first player to be the dealer.
 */
public void resetDealer () {

	players[SOUTH].setDealer(true);
	assignedDealer(players[SOUTH]);
	players[WEST].setDealer(false);
	players[NORTH].setDealer(false);
	players[EAST].setDealer(false);
}

/**
 * Starts the network connection, if possible.
 *
 * @param serverValue	the server name.
 * @param portValue	the port.
 * @return	whether the network started.
 */
public boolean startNetwork (String serverValue, String portValue) {

	if (portValue.length() != 0 && serverValue.length() != 0)
		network = new Network(this, serverValue, Integer.parseInt(portValue));
	else if( portValue.length() != 0)
		network = new Network(this, Integer.parseInt(portValue));
	else if (serverValue.length() != 0)
		network = new Network(this, serverValue);
	else
		network = new Network(this);

	if (network.isValid()) {
		network.start();
		euchre.getGUI().advise("Network Succeeded", "Network connection established.");
		return true;
	} else {
		network = null;
		euchre.getGUI().advise("Network Failed", "Network connection failed.");
		return false;
	}
}

/**
 * Plays a hand.
 */
public void playHand () {

	Enumeration e;
	Card card;
	int i;
	AIPlayer aiPlayer;
	NetworkPlayer networkPlayer;
	HumanPlayer humanPlayer;
	Player player;
	int rc = -1;

	if (waitingForHumanPlayedCard) {
		humanPlayer = (HumanPlayer) players[currentPosition];
		if (! humanPlayer.playCard(SOUTH)) {
			euchre.getGUI().advise("Invalid Card", "Please try again.");
			return;
		}
		if (network != null && ! master) {
			waitingForHumanPlayedCard = false;
			System.out.println("Returning to the beginning.");
			return;
		}
	} else if (waitingForNetworkPlayedCard) {
		System.out.println("Was waiting for network played card.");
		// networkPlayer = (NetworkPlayer) players[currentPosition];
		// networkPlayer.playCard();
	}
    
    	if (! waiting()) {
		newHand();
		System.out.println("Dealing...");
		deal();
		assignedLeader(players[rotate(dealer)]);
	}
	while (trumpCaller == -1 || waitingForHumanDiscardedCard || waitingForNetworkCalled || waitingForNetworkDiscardedCard || waitingForNetworkCalledSuit) {
		establishTrump();
		if (waiting()) System.out.println("Waiting");
		if (waiting()) return;
	}
	System.out.println("Trump established or reentering.");
	// Play five tricks.
	if (! waitingForHumanPlayedCard && ! waitingForNetworkPlayedCard)
		currentTrick = 1;
	while (currentTrick < 6) {
		// Initially the person clockwise from the dealer starts.
		// Thenceforth, the winner starts.
		if (! waitingForHumanPlayedCard && ! waitingForNetworkPlayedCard) {
			if (currentTrick == 1)
				currentPosition = rotate(dealer);
			else
				currentPosition = trickWinner;
			playNum = 0;
		}
		// Play four cards.
		while (playNum < 4) {
			player = players[currentPosition];
			if (waitingForHumanPlayedCard)
				waitingForHumanPlayedCard = false;
			else if (waitingForNetworkPlayedCard)
				waitingForNetworkPlayedCard = false;
			else {
				if (player instanceof AIPlayer) {
					aiPlayer = (AIPlayer) player;
					aiPlayer.playCard(currentPosition);
				} else if (player instanceof HumanPlayer) {
					humanPlayer = (HumanPlayer) player;
					euchre.getGUI().advise("Play Card", "Please play a card.");
					waitingForHumanPlayedCard = true;
					return;
				} else {
					networkPlayer = (NetworkPlayer) player;
					networkPlayer.playCard();
					waitingForNetworkPlayedCard = true;
					return;
				}
			}
			System.out.println("Going to next trick.");
			currentPosition = rotate(currentPosition);
			playNum++;
		}
		determineTrickWinner();
		record.addElement(playedCards);
		playedTrick(); // Causes clients to do the following.
		for (i = 0; i < 4; i++) {
			euchre.getGUI().removeCard(playedCards[i], CENTER);
			playedCards[i] = null;
		}
		currentTrick++;
	}
	euchre.getGUI().removeLeader();
	euchre.getGUI().removeTrump();
	euchre.getGUI().removeTrumpCaller();
	euchre.getGUI().removeTricks();
	euchre.getGUI().removeDealer();
	// Determine hand winner and update scores appropriately.
	processScores();
	// Rotate the dealer.
	players[dealer].setDealer(false);
	i = rotate(dealer);
	players[i].setDealer(true);
	assignedDealer(players[i]);
}

public void setNetworkCard (Card networkCard) {

	this.networkCard = networkCard;
}

public Card getNetworkCard () {

	Card card;

	card = networkCard;
	networkCard = null;
	return card;
}

/**
 * Determines if player's partner is currently winning the trick.
 */
public boolean partnerIsWinning ( ) {

	Vector vector = new Vector();
	Card card, bestCard;
	int i, partnerPosition;
	
	if (playedCards[leader] == null || playedCards[rotate(leader)] ==
	null)
		return false;		// player's team cannot be winning
	for (i = leader; playedCards[i] != null; i = rotate(i)) {
		vector.addElement(playedCards[i]);
	}
	
	partnerPosition = rotate( rotate( currentPosition ) );
	
	card = playedCards[leader];
	bestCard = players[SOUTH].getStrategy().findBestCard(vector, card);
	for (i = leader; playedCards[i] != null; i = rotate(i))
		if (playedCards[i].equals(bestCard))
			break;
	if (i == partnerPosition) 
		return true;
	else
		return false;
}

/**
 * Returns the trick number.
 */
public int getTrickNum () {

	return currentTrick;
}

/**
 * Determines if player's team called trump up.
 */
public boolean weCalledTrump () {
	
	if (currentPosition == trumpCaller)
		return true;	// You called trump
	if (rotate (rotate (currentPosition)) == trumpCaller)
		return true;	// Your partner called trump
	return false;		// They called trump
}

/**
 * Determines the winner of the trick.
 */
public void determineTrickWinner () {

	Card card, bestCard;
	int suit, i, temp;
	Vector vector = new Vector();

	for (i = 0; i < 4; i++)
		vector.addElement(playedCards[i]);
	card = playedCards[leader];
	bestCard = players[SOUTH].getStrategy().findBestCard(vector, card);
	for (i = 0; i < 4; i++)
		if (playedCards[i].equals(bestCard))
			break;
	tricks[i]++;
	trickWinner = i;
	assignedTrickWinner(players[trickWinner]);
	if (trickWinner == NORTH || trickWinner == SOUTH)
		euchre.getGUI().updateOurTricks(tricks[SOUTH] + tricks[NORTH]);
	else
		euchre.getGUI().updateTheirTricks(tricks[EAST] + tricks[WEST]);
	updatedTricks (i, tricks[i]);
	assignedLeader(players[trickWinner]);
	temp = trickWinner + 1;
}

/**
 * Establishes trump.
 */
public void establishTrump () {

	AIPlayer aiPlayer;
	NetworkPlayer networkPlayer;
	HumanPlayer humanPlayer;
	boolean choice = false;
	Player player = null;
	int i, temp, suit;
	Enumeration e;
	Card card;

	System.out.println("Entering establish trump.");

	if (waitingForHumanDiscardedCard || waitingForNetworkDiscardedCard) {
                if (players[currentPosition] instanceof HumanPlayer) {
			System.out.println("Discarding card as human player.");
                        humanPlayer = (HumanPlayer) players[currentPosition];
			humanPlayer.discardCard();
			waitingForHumanDiscardedCard = false;
			return;
                } else if (players[currentPosition] instanceof NetworkPlayer) {
			System.out.println("Discarding card as network player.");
			networkPlayer = (NetworkPlayer) players[currentPosition];                       
			card = findLocalCard(networkPlayer, networkCard);
			if (card == null) {
				System.out.println("Didn't really have that card.");
				System.exit(0);
			}
			networkPlayer.discardCard(card);
			waitingForNetworkDiscardedCard = false;
			return;
		}
	}
	if (! waitingForNetworkCalled && ! waitingForNetworkCalledSuit) {
		currentPosition = rotate(dealer);
		firstGoAroundCount = 0;
	}
	while (firstGoAroundCount < 4) {
		player = players[currentPosition];
		if (player instanceof AIPlayer) {
			aiPlayer = (AIPlayer) player;
			choice = aiPlayer.callItUp();
		} else if (player instanceof HumanPlayer) {
			humanPlayer = (HumanPlayer) player;
			choice = humanPlayer.callItUp();
		} else {
			networkPlayer = (NetworkPlayer) player;
			// Choice will come back to us through the network.
			if (! waitingForNetworkCalled) {
				System.out.println("Need to make a choice.");
				networkPlayer.callItUp();
				waitingForNetworkCalled = true;
				return;
			} else {
				System.out.println("Got an answer: " + networkChoice + ".");
				waitingForNetworkCalled = false;
				choice = networkChoice;
			}
		}
		if (choice)
			break;
		currentPosition = rotate(currentPosition);
		firstGoAroundCount++;
	}
	if (choice) {
		i = trumpCaller + 1;
		euchre.getGUI().blockAdvise("Call it Up", euchre.getGUI().getPlayerName(trumpCaller) + " calls it up.");
		currentPosition = dealer;
		if (players[currentPosition] instanceof AIPlayer) {
			aiPlayer = (AIPlayer) players[currentPosition];
			aiPlayer.pickUp();
		} else if (players[currentPosition] instanceof HumanPlayer) {
			humanPlayer = (HumanPlayer) players[currentPosition];
			humanPlayer.pickUpTrump();
			System.out.println("Got here.");
			euchre.getGUI().advise("Discard", "Please discard a card.");
			waitingForHumanDiscardedCard = true;
			return;
		} else if (players[currentPosition] instanceof NetworkPlayer) {
			networkPlayer = (NetworkPlayer) players[currentPosition];
			networkPlayer.pickUp();
			waitingForNetworkDiscardedCard = true;
			return;
		} else {
			System.out.println("Observer.java: Invalid player type.");
			System.exit(0);
		}
		return;
	}
	// It went around once and nobody called for the dealer to pick it up.
	else {
		if (! waitingForNetworkCalledSuit)
			secondGoAroundCount = 0;
		while (secondGoAroundCount < 4) {
			player = players[currentPosition];
			if (player instanceof AIPlayer) {
				aiPlayer = (AIPlayer) player;
				choice = aiPlayer.callItUp();
			} else if (player instanceof HumanPlayer) {
				humanPlayer = (HumanPlayer) player;
				choice = humanPlayer.callItUp();
			} else {
				networkPlayer = (NetworkPlayer) player;
				if (! waitingForNetworkCalledSuit) {
					networkPlayer.callItUp();
					waitingForNetworkCalledSuit = true;
					System.out.println("Waiting for network called suit.");
					return;
				} else {
					waitingForNetworkCalledSuit = false;
					choice = networkChoice;
				}
			}
			if (choice) break;
			currentPosition = rotate(currentPosition);
			secondGoAroundCount++;
		}
	}
	if (choice) {
		suit = euchre.getTable().getTrumpSuit();
		euchre.getGUI().blockAdvise("Call it Up", euchre.getGUI().getPlayerName(trumpCaller) + " calls " + Card.getSuitDescription(suit));
	}
}

/**
 * Plays the game.
 *
 * @param local	whether this is a local reentry.
 */
public void playGame (boolean local) {

	if (! waiting()) {
		if (network != null) {
			if (! inProgress) {
				// Tell everybody else we're starting the game.
				processOutgoingText("G");
				inProgress = true;
			}
			if (! master)
				return;
		}
		newGame();
	}
	while (euchre.getTable().getOurScore() < 10 && euchre.getTable().getTheirScore() < 10) {
		if (waiting()) {
			if (localWaiting() && ! local)
				return;
			if (networkWaiting() && local)
				return;
		}
		playHand();
		if (waiting() || (network != null && ! master)) {
			System.out.println("Returning.");
			return;
		}
	}
	if (euchre.getTable().getOurScore() >= 10)
		euchre.getGUI().blockAdvise("We Win", "We win!");
	else
		euchre.getGUI().blockAdvise("They Win", "They win.");
	euchre.getGUI().enablePlayGame();
	euchre.getGUI().removeScores();
}

public boolean waiting () {

	if (networkWaiting() || localWaiting())
		return true;
	else
		return false;
}

public boolean networkWaiting () {

	if (waitingForNetworkCalled || waitingForNetworkDiscardedCard || waitingForNetworkCalledSuit || waitingForNetworkPlayedCard)
		return true;
	else
		return false;
}

public boolean localWaiting () {

	if (waitingForHumanDiscardedCard || waitingForHumanPlayedCard)
		return true;
	else
		return false;
}

/**
 * Returns the next player clockwise.
 *
 * @param currentPlayer	the current player.
 * @return      the next player clockwise.
 */
public int rotate(int currentPlayer) {

        return (currentPlayer + 1) % 4;
}

}
