/*
 * Strategy.java
 *
 * Gary Givental
 * John Gotts
 * Saul Rodriguez
 *
 * EECS 481
 *
 * Revision History:
 *
 * 09/09/97 - original implementation.
 * 12/02/97 - early findBest implementation.
 *
 */

import java.util.*;
import java.awt.*;

/**
 * Implements the strategy engine.
 *
 * @version     1.0
 * @author      Gary Givental
 * @author      John Gotts
 * @author      Saul Rodriguez
 */
public class Strategy {

protected Euchre euchre;
protected Observer observer;
protected Player player;

/**
 * Constructs the strategy engine.
 *
 * @param euchre	the card game.
 * @param observer	the game play observer.
 */
public Strategy (Euchre euchre, Observer observer, Player player) {

	this.euchre = euchre;
	this.observer = observer;
	this.player = player;
}


/**
 * Returns the best card based upon the led card.
 *
 * @return	the best card based upon the led card.
 */
public Card findBestCard (Vector cards, Card ledCard) {

	Card highestNonTrump = null;
	Card highestFollowSuit = null;
	Card highestTrump = null;
	int highestTrumpSubRank = Integer.MIN_VALUE;
	int highestFollowRank = Integer.MIN_VALUE;
	int highestSubRank = Integer.MIN_VALUE;
	int trumpSuit, leftSuit, subRank, effectiveSuit;
	Enumeration e;
	Card card;

	// Get suit statistics
	trumpSuit = euchre.getTable().getTrumpSuit();
	leftSuit = findLeftSuit(trumpSuit);
	
	for (e = cards.elements(); e.hasMoreElements();) {
		card = (Card) e.nextElement();
		subRank = card.getSubRank();
		effectiveSuit = card.getSuit();
		// Left and right bauer get special rankings of 6 and 7 of the
		// trump cards, respectively.
		if (card.getValue().compareTo("Jack") == 0) {
			if (effectiveSuit == leftSuit) {
				subRank = 6;
				effectiveSuit = trumpSuit;
			} else if (effectiveSuit == trumpSuit)
				subRank = 7;
		}
		// Card is a trump card or an effective trump card.
		if (effectiveSuit == trumpSuit) {
			if (subRank > highestTrumpSubRank) {
				highestTrumpSubRank = subRank;
				highestTrump = card;
			}
		// Card is not a trump card.
		} else {
			if (ledCard == null) {
				if (subRank > highestSubRank) {
					highestSubRank = subRank;
					highestNonTrump = card;
				}
			}
			else {
				if (card.getSuit() == ledCard.getSuit()) {
					if (subRank > highestFollowRank) {
						highestFollowRank = subRank;
						highestFollowSuit = card;
					}
				}
				if (subRank > highestSubRank) {
					highestSubRank = subRank;
					highestNonTrump = card;
				}
			}
		}
	}
	
	if (highestTrump != null)
		return highestTrump;
	else if (highestFollowSuit != null)
		return highestFollowSuit;
	else 
		return highestNonTrump;
}

/**
 * Returns the best off trump card.
 *
 * @return	the best off trump card.
 */
public Card findBestOff (Vector cards) {

	Card highestOff = null;
	int highestSubRank = Integer.MIN_VALUE;
	int trumpSuit, leftSuit, subRank, effectiveSuit;
	Enumeration e;
	Card card;

	// Get suit statistics
	trumpSuit = euchre.getTable().getTrumpSuit();
	leftSuit = findLeftSuit(trumpSuit);
	
	for (e = cards.elements(); e.hasMoreElements();) {
		card = (Card) e.nextElement();
		subRank = card.getSubRank();
		effectiveSuit = card.getSuit();
		// Left and right bauer get special rankings of 6 and 7 of the
		// trump cards, respectively.
		if (card.getValue().compareTo("Jack") == 0) {
			if (effectiveSuit == leftSuit) 
				effectiveSuit = trumpSuit;
		}
		// Card is not a trump card or an effective trump card.
		if (effectiveSuit != trumpSuit) {
				if (subRank > highestSubRank) {
					highestSubRank = subRank;
					highestOff = card;
				}
		}
			
	}
	
	return highestOff;
}

/**
 * Returns whether the card to be played is legal.
 *
 * @param card	the card to be played.
 *
 * @return	whether the move is legal.
 */
public boolean legalPlay (Card card) {

	Card ledCard;
	Card playCard;
	
	int trumpSuit;
	int leftSuit;
	int ledSuit;
	
	boolean hasLedSuit = false;
	
	// Get trump and next suits...
	trumpSuit = euchre.getTable().getTrumpSuit();
	leftSuit = findLeftSuit(trumpSuit);
	
	// Ask the observer if this the first card to be played.  If it is,
	// unconditionally allow it.
	if (euchre.getObserver().firstPlay() == true)
		return true;
	
	// card is not first, need more info...
	ledCard = euchre.getObserver().getLedCard();
	
	if (ledCard.getSuit() == leftSuit && ledCard.getValue().compareTo("Jack") == 0)
		ledSuit = trumpSuit;
	else
		ledSuit = ledCard.getSuit();
	
	// If they are following suit, allow.
	if (ledSuit == leftSuit) {
		if (card.getSuit() == ledSuit && card.getValue().compareTo("Jack") != 0)
			return true;
	} else if (ledSuit == trumpSuit) {
		if (card.getSuit() == ledSuit || (card.getSuit() == leftSuit && card.getValue().compareTo("Jack") == 0))
			return true;
	} else if (card.getSuit() == ledSuit && ledSuit != trumpSuit && ledSuit != leftSuit)
		return true;
	
	// Player is not following suit.  Must determine if they have any other ledSuits in hand to play.
	// If so, return false.  If not, the play is free to play whatever he/she wants to.
	
	// Determine if any other cards in hand match ledSuit...
	for (int i = 0; i < player.getHand().getCards().size(); i++) {
		playCard = (Card) player.getHand().getCards().elementAt(i);
		if (!playCard.equals(card)) { // Don't check card they're going to play
			if (playCard.getSuit() == ledSuit && !(playCard.getSuit() == leftSuit && playCard.getValue().compareTo("Jack") == 0)) {
				hasLedSuit = true;
				break;
			} else if (ledSuit == trumpSuit && playCard.getSuit() == leftSuit && playCard.getValue().compareTo("Jack") == 0) {
				hasLedSuit = true;
				break;
			}
		}
	}
	
	if (hasLedSuit)		// Player has cards of led suit, and they're not playing them.
		return false;
	
	return true;			// Player does not have cards of led suit, any play is legal.
}

/**
 * Determines if the player (who is currently the dealer) should pick up the
 * card.  Returns the best candidate for a discard or null if the player
 * should not pick up the card.
 *
 * @return	what card the player should discard or null if the player
 *		should not pick up the card.
 */
public Card pickItUp () {
	
	int remove;
	Card card;
	Vector cards;
	
	cards = player.getCards();
	card = euchre.getTable().getTrumpCard();
	// Temporarily add the trump card.
	cards.addElement(card);
	
	//player.printCards();
	if (orderUp()) {
		remove = removeSuit (cards);
		// Remove the trump card.
		cards.removeElementAt (cards.indexOf(card));
		if ( remove >= 0 )
			return (Card) cards.elementAt(remove);
		else
			return (Card) cards.elementAt(removeLowest(cards));
	} else {
		// Remove the trump card.
		cards.removeElement (card);
		return null;
	}
}

/**
 * Returns card to throw away in dealer's hand when forced to pick up trump.
 *
 *@param playerHand		Hand to be analysed.
 *
 *@return				Card to be removed from hand.
 */
public int forceUp ( Vector playerHand ) {

	int remove;
	
	remove = removeSuit ( playerHand );
		
	if ( remove >= 0 ) {
		return remove;
	}
	else
		remove = removeLowest ( playerHand );
	
	return remove;
}

/**
 * Returns suit index that the player should call given the player must call
 * some suit up as trump.
 *
 *@param playerHand 	Hand to be analysed.
 *
 *@return 			Index of suit.
 */
public int forceCall ( Vector playerHand ) {

	int bestSuit = 0;
	int trumpSuit;
	int i;
	
	int bestTrumpPower = 0;
	
	int otherPower = 0;
	int trumpPower = 0;
	int suits = 0;
	
	Card trump;
	
	trump = euchre.getTable().getTrumpCard();
	trumpSuit = trump.getSuit();
	
	for (i = 1; i < 5; i++) {
		if (i != trumpSuit) {
			trumpPower = getTrumpPower ( playerHand, i );
			if (trumpPower > bestTrumpPower) {
				bestTrumpPower = trumpPower;
				bestSuit = i;
			}
		}
	}
	
	return bestSuit;
}

/**
 * Determines if player should order up the dealer.
 * Returns true if player wants card, false otherwise.
 *
 *@return		whether or not to order up dealer.
 */
public boolean orderUp () {
	
	int otherPower = 0;
	int trumpPower = 0;
	int suits = 0;
	Vector cards;
	
	Card trump;
	
	cards = player.getCards();
	//player.printCards();
	trump = euchre.getTable().getTrumpCard();
	
	int trumpSuit;
	
	trumpSuit = trump.getSuit();
	
	return runRules ( cards, trumpSuit );

}

/**
 *
 */
public int callTrump ( Vector playerHand ) {

	int i;
	
	Card trump;
	trump = euchre.getTable().getTrumpCard();
	
	for (i = 1; i < 5; i++) {
		if (i != trump.getSuit() ) {	// make sure you don't call up passed suit
			if ( runRules ( playerHand, i ) )
				break;
		}
	}
	
	if (i == 5)
		return Card.INVALID;
	else
		return i;
}
			
/**
 * Returns true if calling player should order up trump based on cards in hand 
 * Will allow if:
 * 1. Player has both baeurs
 * 2. Player has four trump
 * 3. Player has three strong trump
 * 4. Player has two strong trump and good suit to follow.
 *
 *@param cards			Hand to be analysed
 *@param suit			Suit that will be or is already trump
 *
 *@return 			Whether or not to call trump
 */
public boolean runRules ( Vector cards, int suit ) {

	Card card;
	int otherPower = 0;
	int trumpPower = 0;
	int numSuits = 0, trumpSuit, leftSuit;
	boolean leftBauer = false;
	boolean rightBauer = false;
	
	// Get vital statistics of cards	
	numSuits = getNumSuits (cards);
	
	trumpSuit = suit;
	leftSuit = findLeftSuit (suit);
	
	for (int i = 0; i < cards.size(); i++) {
		card = (Card) cards.elementAt(i);
		if (card.getSuit() == trumpSuit) {
			if (card.getValue().compareTo("Jack") == 0)
				rightBauer = true;
		}
		else if (card.getSuit() == leftSuit &&
		card.getValue().compareTo("Jack") == 0)
			leftBauer = true;
	}
	
	trumpPower = getTrumpPower ( cards, suit );
	otherPower = getOtherPower ( cards, suit );
	
	// Determining logic
		
	if (leftBauer && rightBauer)	// Both bauers in hand -- pick it up
		return true;
	else if (leftBauer || rightBauer) {	// one baeur in hand
		if (numSuits == 2 && trumpPower > 10)	// have 4 trump, call it
			return true;
		else if (trumpPower > 12)			// have three strong trump, call it
			return true;
	}
	else if (trumpPower > 7) {				// have at least two strong trump cards
		if (numSuits == 2 && otherPower > 6)	// good back up, try it
			return true;
		else if (numSuits == 3 && otherPower > 9) // risky, but try it
			return true;
		else if (numSuits == 4 && otherPower > 11) // risky, but try it
			return true;
	}
	
	return false;
}

/**
 * Counts number of different suits a player has, given player's hand.
 *
 *@param playerHand		Hand to be analysed.
 *
 *@return 	number of suits in playerHand.
 */		
private int getNumSuits( Vector playerHand ) {
	
	Card card;
	Integer j;
	
	Vector suit;
	
	suit = new Vector ();
	
	for (int i = 0; i < playerHand.size(); i++) {
		card = (Card) playerHand.elementAt(i);
		j = new Integer( card.getSuit() );
		if ( suit.contains(j) ) ;
		else
			suit.addElement(j);
	}
	
	return suit.size();
}

/**
 * Quantifies the relative power of trump in a player's hand
 *
 *@param playerHand		Hand to be analysed.
 *@param trumpSuit		Suit to be trump if called
 *
 *@return 	"Power Ranking" of trump cards.
 */
private int getTrumpPower ( Vector cards, int trumpSuit ) {
	
	Enumeration e;
	Card card;
	int trumpPower = 0;
	int leftSuit, effectiveSuit, subRank;
		
	// Get suit statistics
	leftSuit = findLeftSuit(trumpSuit);
	
	for (e = cards.elements(); e.hasMoreElements();) {
		card = (Card) e.nextElement();
		subRank = card.getSubRank();
		effectiveSuit = card.getSuit();
		// Left and right bauer get special rankings of 6 and 7 of the
		// trump cards, respectively.
		if (card.getValue().compareTo("Jack") == 0) {
			if (effectiveSuit == leftSuit) {
				trumpPower = trumpPower + 6;
			} else if (effectiveSuit == trumpSuit)
				trumpPower = trumpPower + 7;
		}
		// Card is a trump card or an effective trump card.
		if (effectiveSuit == trumpSuit) 
			trumpPower = trumpPower + subRank;
		// Card is not a trump card.

	}
	
	return trumpPower;
}

/**
 * Quantifies the relative power of non-trump in a player's hand
 *
 *@param playerHand 	Hand to be analysed.
 *
 *@return		"Power Ranking" of non-trump cards.
 */
private int getOtherPower ( Vector cards, int trumpSuit ) {

	Enumeration e;
	Card card;
	int otherPower = 0;
	int leftSuit, effectiveSuit, subRank;
		
	// Get suit statistics
	leftSuit = findLeftSuit(trumpSuit);
	
	for (e = cards.elements(); e.hasMoreElements();) {
		card = (Card) e.nextElement();
		subRank = card.getSubRank();
		effectiveSuit = card.getSuit();
		// Card is a trump card or an effective trump card.
		if (effectiveSuit != trumpSuit) 
			otherPower = otherPower + subRank;
	}
	
	return otherPower;
}

/**
 * Returns the index of the card whose removal would result in the shortening of suit in
 * a player's hand OR returns Card.INVALID if no card has this trait.
 *
 *@param playerHand		Hand to be analysed.
 *
 *@return 	Index of card or -1 if no card exists.
 */
private int removeSuit ( Vector playerHand ) {

	int initNum;
	Card card;
	Card trump;
	int retVal = Card.INVALID;
	int leftSuit, trumpSuit;
	
	trump = euchre.getTable().getTrumpCard();
	trumpSuit = trump.getSuit();
	leftSuit = findLeftSuit(trumpSuit);
	initNum = getNumSuits ( playerHand );
	
	for (int i = 0; i < playerHand.size(); i++) {
		card = (Card) playerHand.elementAt(i);
		if (card.getSuit () != trumpSuit) {
			if (card.getSuit () == leftSuit &&
				card.getValue().compareTo("Jack") == 0 ) ;
			else {
			//card is not trump or left bower and is eligible for removal
			playerHand.removeElementAt(i);
			if ( getNumSuits( playerHand ) < initNum ){
				playerHand.insertElementAt(card, i);
				retVal = i;
				if (retVal == removeLowest( playerHand ) )	
					// found lowest possible card
					return retVal;
			}
			else
				playerHand.insertElementAt(card, i);	//replace card, did no good
			}
		}
	}
	
	return retVal;	// nothing reduces # of suits
}

/**
 * Returns index of lowest ranking card in a hand.
 *
 *@param playerHand 	Hand to be analysed.
 *
 *@return 	Index of lowest ranking card.
 */
private int removeLowest ( Vector playerHand ) {
	
	int lowest;
	int tmp;
	int initVal;
	Card card;
	Card trump;
	
	int remove = -1;
	
	trump = euchre.getTable().getTrumpCard();
	initVal = getOtherPower (playerHand, trump.getSuit() );
	lowest = initVal;
	
	if ( lowest != 0 ) {
		for ( int i = 0; i < playerHand.size(); i++) {
			card = (Card) playerHand.elementAt(i);
			if ( card.getSuit() != trump.getSuit() ) {
				playerHand.removeElementAt(i);
				tmp = getOtherPower (playerHand, trump.getSuit() );
				if ( lowest > initVal - tmp ) {
					lowest = initVal - tmp;
					remove = i;
				}
				playerHand.insertElementAt(card, i);
			}
		}
	}
	else {
		for ( int i = 0; i < playerHand.size(); i++) {
			card = (Card) playerHand.elementAt(i);
			if ( card.getSuit() != trump.getSuit() ) {
				playerHand.removeElementAt(i);
				tmp = getTrumpPower (playerHand, trump.getSuit() );
				if ( lowest > initVal - tmp ) {
					lowest = initVal - tmp;
					remove = i;
				}
				playerHand.insertElementAt(card, i);
			}
		}
	}
	
	return remove;
}

/**
 * Returns the left suit.
 *
 * @param suit	the right suit.
 * @return the left suit.
 */
public int findLeftSuit (int suit) {

	if (suit == Card.SPADE)
		return Card.CLUB;
	else if (suit == Card.CLUB)
		return Card.SPADE;
	else if (suit == Card.HEART)
		return Card.DIAMOND;
	else if (suit == Card.DIAMOND)
		return Card.HEART;
	else
		return Card.NONE;
}

/**
 * Returns the best legal card in the vector.  Returns null if the trick is
 * unwinnable or if the player's partner will unconditionally win due to
 * playing the highest card remaining or the highest card remaining taking into
 * account the players cards.  When this method returns null, findWorst()
 * should be the next method to be called.
 *
 * NOTE: This function has not yet been written.  It's currently an exact
 * duplicate of findWorst; however, the description of how it should be
 * implemented is in place.
 *
 * @param cards		the vector of cards to check.
 * @return			the best card.
 */
public Card findBest (Vector cards /*Player player*/) {
	
	Enumeration e;
	Vector legalCards, playedCards, vector;
	Card card = null;
	Card ledCard, winningCard;
	int trumpSuit, leftSuit, ledSuit;
	boolean partnerWinning = false;
	
	legalCards = new Vector();
	playedCards = new Vector();
	vector = new Vector();
	
	if (euchre.getObserver().firstPlay() == true) {
		
		if (euchre.getObserver().weCalledTrump()) {
			if (euchre.getObserver().getTrickNum() == 1) {
				card = findBestOff(cards);
				if (card == null)
					card = findBestCard(cards, null);
			}
			else
				card = findBestCard(cards, null);	
		}
		else 							
			card = findBestOff(cards);
		
		// line added to make sure a non-null card is returned.
		if (card == null) {
			System.out.println("*****findBest returning null card - avoid");
			return (Card) cards.elementAt(0);
		}
		return card;
	}
	else {	// Not first to play.  Get legal moves.
		for (e = cards.elements();e.hasMoreElements();) {
			card = (Card) e.nextElement();
			if (legalPlay( card ))
				legalCards.addElement(card);
		}
		ledCard = euchre.getObserver().getLedCard();
		playedCards = euchre.getObserver().getPlayedCards();
		winningCard = findBestCard (playedCards, ledCard);
		// Find out if partner currently has it won.
		partnerWinning = euchre.getObserver().partnerIsWinning();
		// If partner is winning, play worst card.
		if (partnerWinning) {
			System.out.println("Partner winning, I'm throwing low");
			card = findWorst(legalCards, true);
			// line added to make sure a non-null card is returned.
			if (card == null) {
				System.out.println("*****findBest calling findWorst returning null card - avoid");
				return (Card) cards.elementAt(0);
			}
			return findWorst(legalCards, true);
		}
		else	{	// Partner is not winning... Play best card.
			card = findBestCard (legalCards, ledCard);
			vector.addElement (winningCard);
			vector.addElement (card);
			if (findBestCard (vector, winningCard) == winningCard) {
				System.out.println("Partner not winning, nothing I can do.");
				card = findWorst(legalCards, true);
				// line added to make sure a non-null card is returned.
				if (card == null) {
					System.out.println("*****findBest calling findBestCard returning null card - avoid");
					return (Card) cards.elementAt(0);
				}
				return card;
			}
			else {
				// line added to make sure a non-null card is returned.
				if (card == null) {
					System.out.println("*****findBest calling findBestCard returning null card - avoid");
					return (Card) cards.elementAt(0);
				}
				System.out.println("Partner not winning, I'm throwing high");
				return card;
			}

		}
	}
}

/**
 * Returns the worst legal card in the vector.  Useful for the dealer who
 * needs to discard and for the player who needs to throw away junk.  This
 * method ignors previously discarded or played cards.
 *
 * @param cards		the vector of cards to check.
 * @param followSuit	whether the player has to follow suit.
 * @return	the worst card.
 */
public Card findWorst (Vector cards, boolean followSuit) {

	Card card;
	Enumeration e;
	Card lowestNonTrump = null;
	Card lowestFollowSuit = null;
	Card lowestTrump = null;
	int trumpSuit, effectiveSuit, leftSuit, i, subRank, ledSuit;
	int lowestTrumpSubRank = Integer.MAX_VALUE;
	int lowestSubRank = Integer.MAX_VALUE;
	int lowestFollowRank = Integer.MAX_VALUE;
	
	trumpSuit = euchre.getTable().getTrumpSuit();
	leftSuit = findLeftSuit(trumpSuit);
	
	ledSuit = euchre.getObserver().getLedSuit();
	if (ledSuit == Card.NONE)
		followSuit = false;
	
	for (e = cards.elements(); e.hasMoreElements();) {
		card = (Card) e.nextElement();
		subRank = card.getSubRank();
		effectiveSuit = card.getSuit();
		// Left and right bauer get special rankings of 6 and 7 of the
		// trump cards, respectively.
		if (card.getValue().compareTo("Jack") == 0) {
			if (effectiveSuit == leftSuit) {
				subRank = 6;
				effectiveSuit = trumpSuit;
			} else if (effectiveSuit == trumpSuit)
				subRank = 7;
		}
		// Card is a trump card or an effective trump card.
		if (effectiveSuit == trumpSuit) {
			if (effectiveSuit == ledSuit) {
				if (subRank < lowestFollowRank) {
					lowestFollowRank = subRank;
					lowestFollowSuit = card;
				}
			}
			if (subRank < lowestTrumpSubRank) {
				lowestTrumpSubRank = subRank;
				lowestTrump = card;
			}
		// Card is not a trump card.
		} else {
			if (effectiveSuit == ledSuit) {
				if (subRank < lowestFollowRank) {
					lowestFollowRank = subRank;
					lowestFollowSuit = card;
				}
			}
			if (subRank < lowestSubRank) {
				lowestSubRank = subRank;
				lowestNonTrump = card;
			}
		}
	}
	
	if (followSuit) {
		if (lowestFollowSuit != null)
			return lowestFollowSuit;
		else if (lowestNonTrump != null)
			return lowestNonTrump;
		else
			return lowestTrump;
	} else {
		if (lowestNonTrump != null)
			return lowestNonTrump;
		else
			return lowestTrump;
	}
}

}

// End of Strategy.java
