/*
 * Server.java
 *
 * Gary Givental
 * John Gotts
 * Saul Rodriguez
 *
 * EECS 481
 *
 * Revision History:
 *
 * 10/13/97 - original implementation.
 * 11/05/97 - integration with improved game and client code.
 *
 * The server runs on the user-specified port or 5000 by default.
 *
 */

import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Implements a server.
 *
 * @version     1.0
 * @author      Gary Givental
 * @author      John Gotts
 * @author      Saul Rodriguez
 */
public class Server {

protected ServerSocket serverSocket;
protected Vector games;
protected int players;

/**
 * Constructs a new server on the specified port.  All streams are initialized
 * to null.
 *
 * @param port		the port number the server will run on.
 * @param players	the number of players per game.
 */
public Server (int port, int players) {

	try {
		serverSocket = new ServerSocket(port);
	} catch (IOException exception) {
		System.out.println("IO exception: " + exception + ".");
		serverSocket = null;
	} catch (Exception exception) {
		System.out.println("Unknown exception: " + exception + ".");
		serverSocket = null;
	}
	this.players = players;
	games = new Vector();
}

/**
 * Accept incoming connections.
 *
 * @return	whether all clients exited successfully.
 */
public boolean acceptConnections () {

	Game game;
	Client client;
	Socket socket;
	InputStream inputStream;
	OutputStream outputStream;
	Enumeration e;
	String errorMessage = null;
	boolean fillIn;

	for (;;) {
		try {
			// Block until we get a new connection.
			socket = serverSocket.accept();
		} catch (IOException exception) {
			errorMessage = new String("IO exception: " + exception + ".");
			break;
		} catch (Exception exception) {
			errorMessage = new String("Unknown exception: " + exception + ".");
			break;
		}
		try {
			inputStream = socket.getInputStream();
		} catch (IOException exception) {
			errorMessage = new String("IO exception: " + exception + ".");
			break;
		} catch (Exception exception) {
			errorMessage = new String("Unknown exception: " + exception + ".");
			break;
		}
		try {
			outputStream = socket.getOutputStream();
		} catch (IOException exception) {
			errorMessage = new String("IO exception: " + exception + ".");
			break;
		} catch (Exception exception) {
			errorMessage = new String("Unknown exception: " + exception + ".");
			break;
		}
		client = new Client(inputStream, outputStream);
		// Add the first game.
		if (games.size() == 0) {
			game = new Game(players);
			game.addClient(client);
		} else {
			// Fill in players as necessary.
			fillIn = false;
			game = null; // Compiler workaround.
			for (e = games.elements(); e.hasMoreElements();) {
				game = (Game) e.nextElement();
				if (! game.isFull()) {
					game.addClient(client);
					fillIn = true;
					break;
				}
			}
			// Start a new game if necessary.
			if (! fillIn) {
				game = new Game(players);
				game.addClient(client);
			}
		}
		client.setGame(game);
		if (! games.contains(game))
			games.addElement(game);
		// Get the client to start servicing its streams.
		client.start();
	}
	System.out.println(errorMessage);
	return false;
}

/**
 * Initializes the server.
 *
 * @param args	the command line arguments.
 */
public static void main(String args[]) {

	Server server;
	int port = 0;
	int players = 0;

	// The server will run on port 5000 if the user doesn't specify a port
	// number.
	if (args.length == 0) {
		port = 5000;
		players = 4;
	} else if (args.length == 1) {
		try {
			port = Integer.parseInt(args[0]);
		} catch (NumberFormatException exception) {
			System.out.println("First argument must be a port number.");
		} catch (Exception exception) {
			System.out.println("Unknown exception: " + exception + ".");
		}
		players = 4;
	} else if (args.length == 2) {
		try {
			port = Integer.parseInt(args[0]);
		} catch (NumberFormatException exception) {
			System.out.println("First argument must be a port number.");
		} catch (Exception exception) {
			System.out.println("Unknown exception: " + exception + ".");
		}
		try {
			players = Integer.parseInt(args[1]);
		} catch (NumberFormatException exception) {
			System.out.println("Second argument must be the number of players in a game.");
		} catch (Exception exception) {
			System.out.println("Unknown exception: " + exception + ".");
		}
	} else
		System.out.println("Invalid number of arguments.");

	if (port != 0 && players != 0) {
		// Start listening for new connections.
		server = new Server(port, players);
		if (server.serverSocket != null) {
			if (! server.acceptConnections()) System.out.println("Server exited unexpectedly.");
		} else
			System.out.println("Unable to open socket on port " + port + ".");
	}
}

}
